from sqlalchemy import create_engine
import pandas as pd
from table_representation import Table
from .sql import SQL


class SQLite(SQL):

    def __init__(self, host, db_name):
        super().__init__()
        self.conn = create_engine(f"sqlite:///{host}{db_name}")

    def get_schemas_list(self) -> list:
        return ['main']

    def get_table_list(self) -> list:
        temp = pd.read_sql("SELECT * FROM sqlite_master", self.conn)
        return list(temp['name'])

    def create_schema(self, schema_name):
        self.conn.execute(f"attach {schema_name} as 'schemaname'")

    def create_table(self, table_data: Table, if_exists_check=False):
        query = f"CREATE TABLE {table_data.table_name}"
        if if_exists_check:
            query += "IF NOT EXISTS"
        query += "("

        for col in table_data.columns:
            query += f"{col.column_name} {col.column_type},"

        # Removing last comma
        query = query[:-1]

        query += ")"
        self.conn.execute(query)

    def remove_table(self, table_name):
        query = f"DROP TABLE {table_name}"
        self.conn.execute(query)

    def get_table_schema_dict(self):
        temp = pd.read_sql("SELECT * FROM sqlite_master", self.conn)
        return {'main': v for v in temp['name']}

    def get_table_data(self, schema, tablename) -> dict:
        temp = pd.read_sql(f"PRAGMA table_info({tablename});", self.conn)[['name', 'type', 'pk']]

        table_data = dict()
        table_data['name'] = tablename
        table_data['permissions'] = []
        table_data['columns'] = {}

        for col in temp['name'].values:
            table_data['columns'][col] = {
                'name': col,
                # list() and type transform are necessary because values by default
                # returns np object and spoils yaml
                'type': str(list(temp[temp['name'] == col]['type'].values)[0]),
                'is_pk': int(list(temp[temp['name'] == col]['pk'].values)[0])
            }

        return table_data



from src.YamlBase.utils import convert_table_data_from_dict_to_table_obj
import yaml


def test_convert_table_data_from_dict_to_table_obj():
    with open('test_base.yml') as f:
        cfg = yaml.load(f, Loader=yaml.Loader)

    test_schema_name = 'main'
    test_table_name = 'table_1'

    converted_table = convert_table_data_from_dict_to_table_obj(cfg['schemas'][test_schema_name][test_table_name],
                                                    test_schema_name)
    assert converted_table.table_name == cfg['schemas'][test_schema_name][test_table_name]['name']
    assert converted_table.schema_name == test_schema_name
    assert converted_table.permissions == cfg['schemas'][test_schema_name][test_table_name]['permissions']

    assert converted_table.columns[0].column_name == \
           cfg['schemas'][test_schema_name][test_table_name]['columns']['column1']['name']

    assert converted_table.columns[0].column_type == \
           cfg['schemas'][test_schema_name][test_table_name]['columns']['column1']['type']

    assert converted_table.columns[0].is_pk == \
           bool(cfg['schemas'][test_schema_name][test_table_name]['columns']['column1']['is_pk'])

    assert converted_table.columns[0].is_sk == \
           bool(cfg['schemas'][test_schema_name][test_table_name]['columns']['column1']['is_sk'])



from sqlalchemy import create_engine
from table_representation import Table

class SQL:

    def __init__(self):
        pass

    def get_schemas_list(self) -> list:
        pass

    def get_table_list(self) -> list:
        pass

    def add_column_to_table(self):
        pass

    def create_schema(self, schema_name):
        pass

    def create_table(self, table_data: Table, if_exists_check=False):
        pass

    def remove_table(self, table_name):
        pass

    def get_table_schema_dict(self):
        pass

    def get_table_data(self, schema, tablename):
        pass


import sys
from sqlalchemy import create_engine
sys.path.extend(['../'])

from dbworkers import SQLiteWorker
from yaml_worker import YamlDataBaseWorker
import pytest


class TestSQLiteworker:

    def setup(self):
        self.cfg = YamlDataBaseWorker("test_base.yml")
        self.sqlite_connector = SQLiteWorker(self.cfg)

    def test_remove_tables(self):
        self.sqlite_connector.insert_new_table(self.cfg.tables_info[0])

        self.sqlite_connector.remove_table(self.cfg.tables_info[0])

        assert not self.sqlite_connector.tables

    def test_insert_new_tables(self):
        self.sqlite_connector.insert_new_table(self.cfg.tables_info[0])
        assert set(self.sqlite_connector.tables) <= set([i.table_name for i in self.cfg.tables_info])

        self.sqlite_connector.remove_table(self.cfg.tables_info[0])

    def test_check_insert_table_possibility(self):
        self.sqlite_connector.insert_new_table(self.cfg.tables_info[0])

        # Not to insert table that already exists
        assert not self.sqlite_connector.check_insert_table_possibility(self.cfg.tables_info[0].table_name)

        assert self.sqlite_connector.check_insert_table_possibility("test_name")

        self.sqlite_connector.remove_table(self.cfg.tables_info[0])

    def test_check_remove_table_possibility(self):
        self.sqlite_connector.insert_new_table(self.cfg.tables_info[0])

        # Not to insert table that already exists
        assert self.sqlite_connector.check_remove_table_possibility(self.cfg.tables_info[0].table_name)

        assert not self.sqlite_connector.check_remove_table_possibility("test_name")

        self.sqlite_connector.remove_table(self.cfg.tables_info[0])

    def test_read_base_information(self):
        self.sqlite_connector.insert_new_table(self.cfg.tables_info[0])
        self.sqlite_connector.read_base_information()

        assert self.sqlite_connector.schemas == ['main']
        assert self.cfg.tables_info[0].table_name in self.sqlite_connector.tables

        self.sqlite_connector.remove_table(self.cfg.tables_info[0])



"""module for working with yaml config"""

from os import path
from exceptions import FileReadError, FieldNotFoundError, ActionConfigSyntaxError
import yaml
from table_representation import Column, Table


class YamlDataBaseWorker:
    """Class checks if data in YAML are correct and then parse it"""

    lvl1_fields_required = ['db_type', 'ip', 'username', 'password', 'db_name']

    def __init__(self, filepath: str):
        self.base_config = self._read_yaml_base_config(filepath)
        self._check_syntax()
        self.db_info, self.tables_info = self._parse_yaml()

    @staticmethod
    def _read_yaml_base_config(filepath: str):
        """
        :param filepath:
        :return:
        """
        if path.exists(filepath):

            try:
                with open(filepath) as file_thread:
                    return yaml.load(file_thread, Loader=yaml.Loader)

            except Exception:
                raise FileReadError(filepath)

        else:
            raise FileNotFoundError(f"File {filepath} not found")

    def _check_syntax(self):

        # Check if lvl_1 fields defined in config
        if not len(
                list(set(self.base_config.keys()) & set(self.lvl1_fields_required))
        ) == len(self.lvl1_fields_required):

            for field in self.lvl1_fields_required:
                if field not in list(self.base_config.keys()):
                    raise FieldNotFoundError(field)

        # Check for errors in schemas input
        if not len(list(self.base_config['schemas'])):
            raise ValueError("Schemas are empty")

        for schema in self.base_config['schemas']:
            if not len(list(self.base_config['schemas'][schema])):
                raise ValueError(f"Schema {schema} is empty")

    def _parse_yaml(self) -> (dict, dict):
        conf_dict = {}

        for field in self.lvl1_fields_required:
            conf_dict[field] = self.base_config[field]

        schemas_info = self.base_config['schemas']

        tables = []

        for schema in schemas_info:

            for table in list(self.base_config['schemas'][schema]):

                columns = []

                for col in list(self.base_config['schemas'][schema][table]['columns']):

                    is_pk = self.base_config['schemas'][schema][table]['columns'][col]['is_pk'] == 1

                    is_sk = self.base_config['schemas'][schema][table]['columns'][col]['is_sk'] == 1
                    sk_link = None

                    if is_sk:
                        sk_link = self.base_config['schemas'][schema][table]['columns'][col]['link']

                    columns.append(
                        Column(
                            self.base_config['schemas'][schema][table]['columns'][col]['name'],
                            self.base_config['schemas'][schema][table]['columns'][col]['type'],
                            is_pk,
                            is_sk,
                            sk_link
                        )
                    )

                permissions = self.base_config['schemas'][schema][table]['permissions']

                tables.append(Table(
                    schema, self.base_config['schemas'][schema][table]['name'], columns, permissions
                ))

        return conf_dict, tables


class YamlActionsWorker:

    allowed_actions = ['insert', 'remove', 'init']

    def __init__(self, filepath: str):
        self.action_table = self.read_yaml(filepath)
        if not self.check_syntax():
            raise ActionConfigSyntaxError

    @staticmethod
    def read_yaml(filepath):
        """
                :param filepath:
                :return:
                """
        if path.exists(filepath):

            try:
                with open(filepath) as file_thread:
                    return yaml.load(file_thread, Loader=yaml.Loader)

            except Exception:
                raise FileReadError(filepath)

        else:
            raise FileNotFoundError(f"File {filepath} not found")

    def check_syntax(self):
        return all([i in self.allowed_actions for i in self.action_table.values()])



"""Main file of yamlbase"""

import argparse
import yaml
import os
from ybase import YBase
from yaml_worker import YamlActionsWorker

import sys


def main():
    sys.path.append('src/yamlbase')

    with open('src/yamlbase/descriptions_cfg.yaml') as f:
        descriptions = yaml.load(f)

    parser = argparse.ArgumentParser(description=descriptions['arg_parse_descr'])
    parser.add_argument('-cfg_db', action='store', dest='cfg_path',
                        help='Path to yaml file with DB config')
    parser.add_argument('-cfg_action', action='store', dest='action',
                        help='Changes that should be made according config')

    args = parser.parse_args()

    base = YBase(args.cfg_path, args.action)
    base.do_actions()


if __name__ == '__main__':
    raise SystemExit(main())


"""Main module fo yamlbase"""

import ybase
from yamlbase import main
# from .dbworkers import *
import dbworkers
# from .sql import *
import sql
import exceptions
import utils
import table_representation
import yaml_worker



from yaml_worker import YamlDataBaseWorker
from sql import SQLite, SQL


class DBWorker:

    def __init__(self, yaml: YamlDataBaseWorker):
        self.db_data = yaml
        self.conn = self.initialize_connection()
        self.db_identifier = f"{self.db_data.db_info['ip']}_{self.db_data.db_info['db_type']}_" \
                             f"{self.db_data.db_info['db_name']}"

    def check_if_used(self):
        pass

    def initialize_connection(self) -> SQL:
        pass

    def read_base_information(self):
        pass

    def check_base_updates(self):
        pass

    def remove_table(self, table_name):
        pass

    def insert_new_table(self, new_table):
        pass

    def generate_config(self):
        pass

    def check_insert_table_possibility(self, new_table_name):
        pass



from .base import DBWorker
from yaml_worker import YamlDataBaseWorker
import os
from sql import SQLite


class SQLiteWorker(DBWorker):
    tables: list
    schemas: list

    def __init__(self, yaml: YamlDataBaseWorker):
        super().__init__(yaml)
        self.db_data = yaml
        self.conn = self.initialize_connection()
        self.db_identifier = f"{self.db_data.db_info['ip']}_{self.db_data.db_info['db_type']}_" \
                             f"{self.db_data.db_info['db_name']}"
        self.read_base_information()

    def check_if_used(self):
        return self.db_identifier in os.listdir('../scanned_bases')

    def initialize_connection(self) -> SQLite:
        return SQLite(self.db_data.db_info['ip'], self.db_data.db_info['db_name'])

    def insert_new_table(self, new_table):
        self.conn.create_table(new_table)
        self.tables = self.conn.get_table_list()

    def remove_table(self, table_obj):
        self.conn.remove_table(table_obj.table_name)

        self.tables = self.conn.get_table_list()

    def read_base_information(self):
        self.schemas = self.conn.get_schemas_list()
        self.tables = self.conn.get_table_list()

    def generate_config(self) -> dict:
        cfg = dict()
        cfg.update(self.db_data.db_info)

        cfg['schemas'] = {}
        cfg['schemas']['main'] = {}
        for table in self.tables:
            table_data = self.conn.get_table_data('main', table)
            cfg['schemas']['main'].update(table_data)

        return cfg

    def check_insert_table_possibility(self, new_table_name) -> bool:
        """
        :param new_table_name:name of new table
        :return: true if inserting is possible
        """
        # Check if table already exists
        for table in self.tables:
            if table == new_table_name:
                return False

        # Check if all tables that exists in base - defined in config
        for table in self.conn.get_table_schema_dict().values():
            if table not in [i.table_name for i in self.db_data.tables_info]:
                return False

        return True

    def check_remove_table_possibility(self, table_name_to_remove) -> bool:
        """
        :param table_name_to_remove: name of table to check
        :return: if table can be removed
        """

        return table_name_to_remove in self.tables
